///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hdate;

import 'HVal.dart';
import 'HTime.dart';
import 'HTimeZone.dart';
import 'HDateTime.dart';
import 'ParseError.dart';
import 'io/HZincReader.dart';
import 'package:quiver/core.dart';

/// HDate models a date (day in year) tag value.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HDate extends HVal {

  /** Construct from basic fields */
  factory HDate(int year, int month, int day) {
    if (year < 1900) throw new FormatException("Invalid year", year);
    if (month < 1 || month > 12) throw new FormatException("Invalid month", month);
    if (day < 1 || day > 31) throw new FormatException("Invalid day", day);
    return new HDate._(year, month, day);
  }

  /** Construct from Dart DateTime instance */
  factory HDate.fromDateTime(DateTime dt) => new HDate(dt.year, dt.month, dt.day);

  /** Parse from string fomat "YYYY-MM-DD" or raise ParseException */
  factory HDate.fromString(String s) {
    HVal val = new HZincReader(s).readScalar();
    if (val is HDate) return val;
    throw new ParseError(s);
  }

  /** Get HDate for current time in default timezone */
  static HDate today() {
    return new HDateTime.now().date;
  }

  HDate._(this.year, this.month, this.day);

  /** Four digit year such as 2011 */
  final int year;

  /** Month as 1-12 (Jan is 1, Dec is 12) */
  final int month;

  /** Day of month as 1-31 */
  final int day;

  /** Hash is based on year, month, day */
  int _hashCode;
  int get hashCode {
    if (_hashCode == null) {
      _hashCode = hash3(year, month, day);
    }
    return _hashCode;
  }

  /** Equals is based on year, month, day */
  bool operator ==(Object that) {
    if (that is! HDate) return false;
    HDate x = that as HDate;
    return year == x.year && month == x.month && day == x.day;
  }

  /** Return sort order as negative, 0, or positive */
  @override
  int compareTo(Object that) {
    HDate x = that as HDate;
    if (year < x.year)   return -1; else if (year > x.year)   return 1;
    if (month < x.month) return -1; else if (month > x.month) return 1;
    if (day < x.day)     return -1; else if (day > x.day)     return 1;
    return 0;
  }

  /** Encode as "YYYY-MM-DD" */
  String _toZinc;
  String toZinc() {
    if (_toZinc == null) {
      StringBuffer s = new StringBuffer();
      encodeForZinc(s);
      _toZinc = s.toString();
    }
    return _toZinc;
  }

  /** Package private implementation shared with HDateTime */
  void encodeForZinc(StringBuffer s) {
    s.write(year);
    s.write('-');

    if (month < 10)
      s.write('0');
    s.write(month); s.write('-');

    if (day < 10)
      s.write('0');
    s.write(day);
  }

  /** Convert this date into HDateTime for midnight in given timezone. */
  HDateTime midnight(HTimeZone tz) {
    return new HDateTime(this, HTime.MIDNIGHT, tz);
  }

  /** Return date in future given number of days */
  HDate plusDays(int numDays) {
    if (numDays == 0) return this;
    if (numDays < 0) return minusDays(-numDays);
    int year  = this.year;
    int month = this.month;
    int day   = this.day;

    for (; numDays > 0; --numDays) {
      day++;
      if (day > daysInMonth(year, month)) {
        day = 1;
        month++;
        if (month > 12) { month = 1; year++; }
      }
    }
    return new HDate(year, month, day);
  }

  /** Return date in past given number of days */
  HDate minusDays(int numDays) {
    if (numDays == 0) return this;
    if (numDays < 0) return plusDays(-numDays);
    int year  = this.year;
    int month = this.month;
    int day   = this.day;

    for (; numDays > 0; --numDays) {
      day--;
      if (day <= 0) {
        month--;
        if (month < 1) { month = 12; year--; }
        day = daysInMonth(year, month);
      }
    }
    return new HDate(year, month, day);
  }

  /** Return if given year a leap year */
  static bool isLeapYear(int year) => ((year & 3) == 0) && ((year % 100 != 0) || (year % 400 == 0));

  /** Return number of days in given year (xxxx) and month (1-12) */
  static int daysInMonth(int year, int mon) => isLeapYear(year) ? daysInMonLeap[mon] : daysInMon[mon];

  /** Return day of week: Sunday is 1, Saturday is 7 */
  int weekday() => new DateTime(year, month, day).weekday;

  static final List<int> daysInMon     = new List<int>.from([ -1, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]);
  static final List<int> daysInMonLeap = new List<int>.from([ -1, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]);
}