///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hdate_time;

import 'HVal.dart';
import 'HDate.dart';
import 'HTime.dart';
import 'HTimeZone.dart';
import 'ParseError.dart';
import 'io/HZincReader.dart';
import 'package:quiver/core.dart';

/// HDateTime models a timestamp with a specific timezone.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HDateTime extends HVal {

  factory HDateTime.fromValues(int year, int month, int day, int hour, int min, int sec, HTimeZone tz) {
    HDate date = new HDate(year, month, day);
    HTime time = new HTime(hour, min, sec);
    if (tz == null) tz = HTimeZone.DEFAULT;
    return new HDateTime._(date, time, tz);
  }

  factory HDateTime.fromMillis(int ms, [HTimeZone tz]) {
    return new HDateTime.fromDateTime(new DateTime.fromMillisecondsSinceEpoch(ms, isUtc: true), tz);
  }

  factory HDateTime.fromDateTime(DateTime dt, [HTimeZone tz]) {
    if (tz == null) tz = HTimeZone.DEFAULT;
    DateTime fixedTime = dt.add(new Duration(seconds: tz.offset));
    return new HDateTime(new HDate.fromDateTime(fixedTime), new HTime.fromDateTime(fixedTime), tz);
  }

  factory HDateTime.fromString(String s) {
    HVal val = new HZincReader(s).readScalar();
    if (val is HDateTime) return val;
    throw new ParseError(s);
  }

  factory HDateTime.now([HTimeZone tz]) => new HDateTime.fromDateTime(new DateTime.now(), tz);

  factory HDateTime(HDate date, HTime time, [HTimeZone tz]) {
    if (date == null || time == null) throw new ArgumentError('data and time must not be null');
    if (tz == null) tz = HTimeZone.DEFAULT;
    return new HDateTime._(date, time, tz);
  }

  HDateTime._(this.date, this.time, this.tz);

  /** Date component of the timestamp */
  final HDate date;

  /** Time component of the timestamp */
  final HTime time;

  /** Offset in seconds from UTC including DST offset */
//  final int tzOffset;

  /** Timezone as Olson database city name */
  final HTimeZone tz;

  /** Get this date time as Java milliseconds since epoch */
  int _millis;
  int millis() {
    if (_millis == null) {
      DateTime utcTime = new DateTime.utc(date.year, date.month, date.day, time.hour, time.min, time.sec, time.ms);
      _millis = utcTime.subtract(new Duration(seconds: tz.offset)).millisecondsSinceEpoch;
    }
    return _millis;
  }

  /** Hash is based on date, time, tzOffset, and (NOT) tz */
  int _hashCode;
  int get hashCode {
    if (_hashCode == null) {
      _hashCode = hash3(date.hashCode, time.hashCode, tz.hashCode);
    }
    return _hashCode;
  }

  bool operator ==(Object that) {
    if (that is! HDateTime) return false;
    HDateTime x = that;
    return date == x.date && time == x.time && tz == x.tz && tz.name == x.tz.name;
  }

  /** Comparison based on millis. */
  int compareTo(Object that) {
    int thisMillis = this.millis();
    int thatMillis = (that as HDateTime).millis();
    if (thisMillis < thatMillis) return -1;
    else if (thisMillis > thatMillis) return 1;
    return 0;
  }

  /** Encode as "YYYY-MM-DD'T'hh:mm:ss.FFFz zzzz" */
  String toZinc() {
    StringBuffer s = new StringBuffer();
    date.encodeForZinc(s);
    s.write('T');
    time.encodeForZinc(s);
    if (tz.offset == 0)
      s.write('Z');
    else {
      int offset = this.tz.offset; //seconds
      if (offset < 0) {
        s.write('-');
        offset = -offset;
      } else {
        s.write('+');
      }

      int zh = offset ~/ 3600;
      int zm = (offset % 3600) ~/ 60;
      if (zh < 10) s.write('0'); s.write(zh); s.write(':');
      if (zm < 10) s.write('0'); s.write(zm);
    }

    s.write(' ');
    s.write(tz);
    return s.toString();
  }
}