///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hgrid;

import 'HVal.dart';

import 'HRef.dart';
import 'HDateTime.dart';
import 'HStr.dart';
import 'HNum.dart';
import 'HBool.dart';
import 'HMarker.dart';
import 'io/HZincWriter.dart';
import 'UnknownNameError.dart';
import 'str-utils.dart';
import 'package:quiver/core.dart';

part 'HDict.dart';
part 'HRow.dart';
part 'HCol.dart';
part 'HHisItem.dart';

/// HGrid is an immutable two dimension data structure of cols and rows.
/// Use HGridBuilder to construct a HGrid instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/Grids)
class HGrid {
  /** Empty grid with one column called "empty" and zero rows */
  static final HGrid EMPTY = new HGrid(
      HDict.EMPTY,
      new List<HCol>.from([new HCol(0, "empty", HDict.EMPTY)]),
      new List()
  );

  HGrid(HDict this._meta, List<HCol> this._cols, List<List<HVal>> rowList)
          : _rows = new List<HRow>(),
          _colsByName = new Map<String,HCol>()
  {
    if (meta == null) throw new ArgumentError("metadata cannot be null");

    for (int i=0; i<rowList.length; ++i) {
      List<HVal> cells = rowList[i];
      if (_cols.length != cells.length)
        throw new ArgumentError("Row cells size != cols size");

      _rows.add(new HRow(this, cells));
    }

    for (int i=0; i<_cols.length; ++i) {
      HCol col = _cols[i];
      String colName = col.name;
      if (_colsByName.containsKey(colName))
        throw new ArgumentError("Duplicate col name: $colName");

      _colsByName[colName] = col;
    }
  }


//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /** Return grid level meta */
  HDict get meta => _meta;

  List<HCol> get cols => _cols;

  /** Error grid have the meta.err marker tag */
  bool get isErr => meta.has("err");

  /** Return if number of rows is zero */
  bool get isEmpty => numRows == 0;

  /** Return number of rows */
  int get numRows => _rows.length;

  /** Get a row by its zero based index */
  HRow row(int i) => _rows[i];

  /** Get number of columns  */
  int get numCols => _cols.length;

  // Get a column by name.  If not found and checked if false then
  // return null, otherwise throw UnknownNameException
  HCol colByName(String name, [bool checked = true]) {
    HCol col = _colsByName[name];
    if (col != null) return col;
    if (checked) throw new UnknownNameError(name);
    return null;
  }

  // Get a column by its index
  HCol colByIndex(int index, [bool checked = true]) => _cols[index];

  /** Create iteratator to walk each row */
  Iterator<HRow> iterator() => new _GridIterator(_rows);


//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  /** Debug dump - this is Zinc right now. */
  void dump(StringBuffer out) {
    out.write(HZincWriter.gridToString(this));
  }

//////////////////////////////////////////////////////////////////////////
// Rows
//////////////////////////////////////////////////////////////////////////

  final List<HRow> _rows;
  final List<HCol> _cols;
  final HDict _meta;
  final Map<String,HCol> _colsByName;

}

//////////////////////////////////////////////////////////////////////////
// GridIterator
//////////////////////////////////////////////////////////////////////////

class _GridIterator implements Iterator {

  _GridIterator(this._rows);

  bool moveNext() {
    if (_pos < _rows.length) _pos++;

    if (_pos < _rows.length) {
      _current = _rows[_pos];
      return true;
    } else
      return false;
  }

  int _pos = -1;
  final List<HRow> _rows;

  HRow _current;
  HRow get current => _current;

}