///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hgrid_builder;

import 'HVal.dart';
import 'HGrid.dart';
import 'HDictBuilder.dart';

/// HGridBuilder is used to construct an immutable HGrid instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/Grids)
class HGridBuilder {

  // todo port follows static methods
//  /** Convenience to build one row grid from HDict. */
//  public static HGrid dictToGrid(HDict dict)
//  {
//    HGridBuilder b = new HGridBuilder();
//    Iterator it = dict.iterator();
//    ArrayList cells = new ArrayList();
//    while (it.hasNext())
//    {
//      Map.Entry entry = (Map.Entry)it.next();
//      String name = (String)entry.getKey();
//      HVal val = (HVal)entry.getValue();
//      b.addCol(name);
//      cells.add(val);
//    }
//    b.rows.add(cells.toArray(new HVal[cells.size()]));
//    return b.toGrid();
//  }
//
//  /** Convenience to build grid from array of HDict.
//      Any null entry will be row of all null cells. */
//  public static HGrid dictsToGrid(HDict[] dicts)
//  {
//  return dictsToGrid(HDict.EMPTY, dicts);
//  }

  /** Convenience to build grid from array of HDict.
      Any null entry will be row of all null cells. */
  static HGrid dictsToGrid(HDict meta, List<HDict> dicts) {
    if (dicts.length == 0) return new HGrid(
        meta,
        [new HCol(0, "empty", HDict.EMPTY)],
        []
    );

    HGridBuilder b = new HGridBuilder();
    b.meta.addDict(meta);

    // collect column names
    Map colsByName = new Map();
    for (int i=0; i<dicts.length; ++i) {
      HDict dict = dicts[i];
      if (dict == null) continue;
      Iterator<DictEntry> it = dict.iterator();
      while (it.moveNext()) {
        DictEntry entry = it.current;
        String name = entry.key;
        if (!colsByName.containsKey(name)) {
          colsByName[name] = name;
          b.addCol(name);
        }
      }
    }

    // if all dicts were null, handle special case
    // by creating a dummy column
    if (colsByName.length == 0) {
      colsByName["empty"] = "empty";
      b.addCol("empty");
    }

    // now map rows
    int numCols = b._cols.length;
    for (int ri=0; ri<dicts.length; ++ri)
    {
      HDict dict = dicts[ri];
      List<HVal> cells = new List<HVal>(numCols);
      for (int ci=0; ci<numCols; ++ci)
      {
        if (dict == null)
          cells[ci] = null;
        else
          cells[ci] = dict.get(b._cols.elementAt(ci).name, checked:false);
      }
      b._rows.add(cells);
    }

    return b.toGrid();
  }

  //todo
//  /** Convenience to build an error grid from exception */
//  public static HGrid errToGrid(Throwable e)
//  {
//    // Java sucks
//    StringWriter sout = new StringWriter();
//    PrintWriter pout = new PrintWriter(sout);
//    e.printStackTrace(pout);
//    pout.flush();
//    String trace = sout.toString();
//    StringBuffer temp = new StringBuffer(trace.length());
//    for (int i=0; i<trace.length(); ++i)
//    {
//      int ch = trace.charAt(i);
//      if (ch == '\t') temp.append("  ");
//      else if (ch != '\r') temp.append((char)ch);
//    }
//    trace = temp.toString();
//
//    HGridBuilder b = new HGridBuilder();
//    b.meta().add("err")
//    .add("dis", e.toString())
//    .add("errTrace", trace);
//    b.addCol("empty");
//    return b.toGrid();
//  }

  /** Convenience to build grid from array of HHisItem */
  static HGrid hisItemsToGrid(HDict meta, List<HHisItem> items) {
    HGridBuilder b = new HGridBuilder();
    b.meta.addDict(meta);
    b.addCol("ts");
    b.addCol("val");
    for (int i=0; i<items.length; ++i) {
      b._rows.add([items[i].ts, items[i].val]);
    }
    return b.toGrid();
  }

  HGridBuilder() {

  }

//////////////////////////////////////////////////////////////////////////
// Building
//////////////////////////////////////////////////////////////////////////

  /** Get the builder for the grid meta map */
  HDictBuilder get meta => _meta;

  /** Add new column and return builder for column metadata.
      Columns cannot be added after adding the first row. */
  HDictBuilder addCol(String name) {
    if (_rows.length > 0)
      throw new StateError("Cannot add cols after rows have been added");
    if (!HDict.isTagName(name))
      throw new ArgumentError("Invalid column name: $name");
    BCol col = new BCol(name);
    _cols.add(col);
    return col.meta;
  }

  /** Add new row with array of cells which correspond to column
      order.  Return this. */
  HGridBuilder addRow(List<HVal> cells) {
    if (_cols.length != cells.length)
      throw new ArgumentError("Row cells size != cols size");
    _rows.add(new List<HVal>.from(cells));
    return this;
  }

  HGrid toGrid() {
    // meta
    HDict meta = this._meta.toDict();

    // cols
    List<HCol> hcols = new List<HCol>();
    for (int i=0; i<_cols.length; ++i) {
      BCol bc = _cols[i];
      hcols.add(new HCol(i, bc.name, bc.meta.toDict()));
    }

    // let HGrid constructor do the rest...
    return new HGrid(meta, hcols, _rows);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final HDictBuilder _meta = new HDictBuilder();
  final List<BCol> _cols = new List<BCol>();
  final List<List<HVal>> _rows = new List<List<HVal>>();

}

//////////////////////////////////////////////////////////////////////////
// BCol
//////////////////////////////////////////////////////////////////////////

class BCol {
  BCol(this.name);

  final String name;
  final HDictBuilder meta = new HDictBuilder();
}