///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hnum;

import 'HVal.dart';
import 'str-utils.dart';
import 'package:intl/intl.dart';

/// HNum wraps a 64-bit floating point number and optional unit name.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HNum extends HVal {

  factory HNum (num val, [String unit]) {
    if (val != null) val = val.toDouble();
    if (val == 0 && unit == null) return ZERO;
    return new HNum._(val, unit);
  }

  /** Singleton value for zero */
  static final HNum ZERO = new HNum._(0.0, null);

  /** Singleton value for positive infinity "Inf" */
  static final HNum POS_INF = new HNum._(double.INFINITY, null);

  /** Singleton value for negative infinity "-Inf" */
  static final HNum NEG_INF = new HNum._(double.NEGATIVE_INFINITY, null);

  /** Singleton value for not-a-number "NaN" */
  static final HNum NaN = new HNum._(double.NAN, null);

  /** Private constructor */
  HNum._(this.val, this.unit) {
    if (!isUnitName(unit)) throw new ArgumentError("Invalid unit name: $unit");
  }

  /** Double scalar value */
  final double val;

  /** Unit name or null */
  final String unit;

  /** Hash code is based on val, unit */
  int _hashcode;
  int get hashCode {
    if (_hashcode == null) {
      _hashcode = val.hashCode;
      if (unit != null) _hashcode ^= unit.hashCode;
    }
    return _hashcode;
  }

  /** Equals is based on val, unit (NaN == NaN) */
  @override
  bool operator ==(Object that) {
    if (that is! HNum) return false;
    HNum x = that as HNum;
    if (val != null && x.val != null && val.isNaN) return x.val.isNaN;
    if (val != x.val) return false;
    if (unit == null) return x.unit == null;
    if (x.unit == null) return false;
    return unit == x.unit;
  }

  /** Return sort order as negative, 0, or positive */
  @override
  int compareTo(Object that) {
    double thatVal = (that as HNum).val;
    if (this.val < thatVal) return -1;
    if (this.val == thatVal) return 0;
    return 1;
  }

  /** Encode as floating value followed by optional unit string */
  String _toZing;
  String toZinc() {
    if (_toZing == null) {
      StringBuffer s = new StringBuffer();
      if (val == double.INFINITY) {
        s.write("INF");
      } else if (val == double.NEGATIVE_INFINITY) {
        s.write("-INF");
      } else if (val.isNaN) {
        s.write("NaN");
      } else {
        //don't let fractions
        if (val.abs() > 1.0)
          s.write(new NumberFormat('#0.####', 'en_US').format(val));
        else
          s.write(val);

        if (unit != null) s.write(unit);
      }
      _toZing = s.toString();
    }
    return _toZing;
  }

  /**
   * Get this number as a duration of milliseconds.
   * Raise IllegalStateException if the unit is not a duration unit.
   */
  int millis() {
    String u = this.unit;
    if (u == null) u = "null";
    if (u == "ms"  || u == "millisecond") return val.toInt();
    if (u == "s"   || u == "sec")         return (val*1000.0).toInt();
    if (u == "min" || u == "minute")      return (val*1000.0*60.0).toInt();
    if (u == "h"   || u == "hr")          return (val*1000.0*60.0*60.0).toInt();
    throw new FormatException("Invalid duration unit: $u");
  }

  /**
   * Return true if the given string is null or contains only valid unit
   * chars.  If the unit name contains invalid chars return false.  This
   * method does *not* check that the unit name is part of the standard
   * unit database.
   */
  static bool isUnitName(String unit) {
    if (unit == null) return true;
    if (unit.length == 0) return false;

    return unit.codeUnits.every((int c) {
      return (c < 128 && !_unitChars[c])? false : true;
    });
  }

  static final List<bool> _unitChars = (){
    List<bool> result = new List<bool>.filled(128, false);
    for (int i=a_CHAR; i<=z_CHAR; ++i) result[i] = true;
    for (int i=A_CHAR; i<=Z_CHAR; ++i) result[i] = true;
    result[UNDERSCORE_CHAR] = true;
    result[DOLLAR_CHAR] = true;
    result[PERCENT_CHAR] = true;
    result[SLASH_CHAR] = true;

    return result;
  }();

}