///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.htime;

import 'HVal.dart';
import 'ParseError.dart';
import 'io/HZincReader.dart';
import 'package:quiver/core.dart';

/// HTime models a time of day tag value.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HTime extends HVal {

  /** Construct with all fields */
  factory HTime(int hour, int min, [int sec = 0, int ms = 0]) {
    if (hour < 0 || hour > 23)  throw new FormatException("Invalid hour", hour);
    if (min  < 0 || min  > 59)  throw new FormatException("Invalid min", min);
    if (sec  < 0 || sec  > 59)  throw new FormatException("Invalid sec", sec);
    if (ms   < 0 || ms   > 999) throw new FormatException("Invalid ms", ms);
    return new HTime._(hour, min, sec, ms);
  }

  /** Construct from Dart DateTime instance */
  factory HTime.fromDateTime(DateTime dt) => new HTime(dt.hour, dt.minute, dt.second, dt.millisecond);

  /** Parse from string fomat "hh:mm:ss.FF" or raise ParseException */
  factory HTime.fromString(String s) {
    HVal val = new HZincReader(s).readScalar();
    if (val is HTime)
      return val;
    else
      throw new ParseError(s);
  }

  /** Singleton for midnight 00:00 */
  static final HTime MIDNIGHT = new HTime._(0, 0, 0, 0);

  HTime._(this.hour, this.min, this.sec, this.ms);

  /** Hour of day as 0-23 */
  final int hour;

  /** Minute of hour as 0-59 */
  final int min;

  /** Second of minute as 0-59 */
  final int sec;

  /** Fractional seconds in milliseconds 0-999 */
  final int ms;

  /** Hash is based on hour, min, sec, ms */
  int _hashCode;
  int get hashCode {
    if (_hashCode == null) {
      _hashCode = hash4(hour, min, sec, ms);
    }
    return _hashCode;
  }

  /** Equals is based on year, month, day */
  bool operator ==(Object that) {
    if (that is! HTime) return false;
    HTime x = that as HTime;
    return hour == x.hour && min == x.min && sec == x.sec && ms == x.ms;
  }

  /** Return sort order as negative, 0, or positive */
  @override
  int compareTo(Object that) {
    HTime x = that as HTime;
    if (hour < x.hour) return -1; else if (hour > x.hour) return 1;
    if (min < x.min)   return -1; else if (min > x.min)   return 1;
    if (sec < x.sec)   return -1; else if (sec > x.sec)   return 1;
    if (ms < x.ms)     return -1; else if (ms > x.ms)     return 1;

    return 0;
  }

  /** Encode as "hh:mm:ss.FFF" */
  String _toZinc;
  String toZinc() {
    if (_toZinc == null) {
      StringBuffer s = new StringBuffer();
      encodeForZinc(s);
      _toZinc = s.toString();
    }
    return _toZinc;
  }

  /** Package private implementation shared with HDateTime */
  void encodeForZinc(StringBuffer s) {
    if (hour < 10) s.write('0'); s.write(hour); s.write(':');
    if (min  < 10) s.write('0'); s.write(min); s.write(':');
    if (sec  < 10) s.write('0'); s.write(sec);
    if (ms != 0) {
      s.write('.');
      if (ms < 10) s.write('0');
      if (ms < 100) s.write('0');
      s.write(ms);
    }
  }
}