///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hzincreader;

import '../HVal.dart';
import '../HMarker.dart';
import '../HNum.dart';
import '../HBool.dart';
import '../HDate.dart';
import '../HUri.dart';
import '../HRef.dart';
import '../HCoord.dart';
import '../HBin.dart';
import '../HStr.dart';
import '../HDateTime.dart';
import '../HTimeZone.dart';
import '../HTime.dart';
import '../HGrid.dart';
import '../HGridBuilder.dart';
import '../HDictBuilder.dart';
import 'StringReader.dart';
import '../str-utils.dart';
import '../ParseError.dart';

/// HZincReader reads grids using the Zinc format.
///
/// See [Project Haystack](http://project-haystack.org/doc/Zinc)
class HZincReader {

  HZincReader(String s) :_in = new StringReader(s) {
    _init();
  }

  void _init() {
    consume();
    consume();
  }

//////////////////////////////////////////////////////////////////////////
// HGridReader
//////////////////////////////////////////////////////////////////////////

  /** Read grid from the stream. */
  HGrid readGrid() {
    HGridBuilder b = new HGridBuilder();

    // meta line
    _readVer();
    _readMeta(b.meta);
    _consumeNewline();

    // read cols
    int numCols = 0;
    while (true) {
      String name = _readId();
      _skipSpace();
      numCols++;
      _readMeta(b.addCol(name));
      if (_cur != COMMA_CHAR) break;
      consume();
      _skipSpace();
    }
    _consumeNewline();

    // rows
    while (_cur != SLASH_N_CHAR && _cur > 0) {
      List<HVal> cells = new List<HVal>();

      for (int i=0; i<numCols; ++i) {
        _skipSpace();
        if (_cur != COMMA_CHAR && _cur != SLASH_N_CHAR)
          cells.add(_readVal());
        else //fixme may be it's wrong. It's may way to fix it (kelegorm)
          cells.add(null);
        _skipSpace();
        if (i+1 < numCols) {
          if (_cur != COMMA_CHAR)
            throw _errChar("Expecting comma in row");
          consume();
        }
      }

      _consumeNewline();
      b.addRow(cells);
    }
    if (_cur == SLASH_N_CHAR) _consumeNewline();

    return b.toGrid();
  }

  /** Read list of grids from the stream. */
  List<HGrid> readGrids() {
    List<HGrid> acc = new List<HGrid>();
    while (_cur > 0) acc.add(readGrid());
    return acc;
  }


  /** Read set of name/value tags as dictionary */
  HDict readDict() {
    HDictBuilder b = new HDictBuilder();
    _readMeta(b);
    if (_cur >= 0) throw _errChar("Expected end of stream");
    return b.toDict();
  }

//////////////////////////////////////////////////////////////////////////
// Implementation
//////////////////////////////////////////////////////////////////////////

  void _readVer() {
    String id = _readId();
    if (id != ("ver")) throw _err("Expecting zinc header 'ver:2.0', not '$id'");
    if (_cur != COLON_CHAR) throw _err("Expecting ':' colon");
    consume();
    String ver = _readStrLiteral();
    if (ver == "2.0")
      _version = 2;
    else
      throw _err("Unsupported zinc version: " + ver);
    _skipSpace();
  }

  void _readMeta(HDictBuilder b) {
    // parse pairs
    while (_isIdStart(_cur))
    {
      // name
      String name = _readId();

      // marker or :val
      HVal val = HMarker.VAL;
      _skipSpace();
      if (_cur == COLON_CHAR)
      {
        consume();
        _skipSpace();
        val = _readVal();
        _skipSpace();
      }
      b.add(name, val);
      _skipSpace();
    }
  }

  String _readId() {
    if (!_isIdStart(_cur)) throw _errChar("Invalid name start char");
    StringBuffer s = new StringBuffer();
    while (_isId(_cur)) {
      s.writeCharCode(_cur);
      consume();
    }
    return s.toString();
  }

//////////////////////////////////////////////////////////////////////////
// HVals
//////////////////////////////////////////////////////////////////////////

  /** Read scalar value. */
  HVal readScalar() {
    HVal val = _readVal();
    if (_cur >= 0) throw _errChar("Expected end of stream");
    return val;
  }

  /** Read a single scalar value from the stream. */
  HVal _readVal() {
    if (_isDigit(_cur)) return _readNumVal();
    if (_isAlpha(_cur)) return _readWordVal();
    switch (_cur) {
      case DOG_CHAR: return _readRefVal();
      case DOUBLE_QUOTE_CHAR: return _readStrVal();
      case GRAVE_ACCENT_CHAR: return _readUriVal();
      case MINUS_CHAR:
        if (_peek == I_CHAR) return _readWordVal();
        return _readNumVal();
      default:  throw _errChar("Unexpected char for start of value");
    }
  }

  HVal _readWordVal() {
    // read into string
    StringBuffer s = new StringBuffer();
    do {
      s.writeCharCode(_cur);
      consume();
    } while (_isAlpha(_cur));

    String word = s.toString();

    // match identifier
    if (_isFilter) {
      if (word == "true")  return HBool.TRUE;
      if (word == "false") return HBool.FALSE;
    } else {
      if (word == "N")   return null;
      if (word == "M")   return HMarker.VAL;
      if (word == "R")   return new HStr("_remove_");
      if (word == "T")   return HBool.TRUE;
      if (word == "F")   return HBool.FALSE;
      if (word == "Bin") return _readBinVal();
      if (word == "C")   return _readCoordVal();
    }
    if (word == "NaN") return HNum.NaN;
    if (word == "INF") return HNum.POS_INF;
    if (word == "-INF") return HNum.NEG_INF;
    throw _err("Unknown value identifier: " + word);
  }

  HVal _readBinVal() {
    if (_cur < 0) throw _err("Expected '(' after Bin");
    consume();
    StringBuffer s = new StringBuffer();
    while (_cur != RIGHT_CURLY_CHAR) {
      if (_cur < 0) throw _err("Unexpected end of bin literal");
      if (_cur == SLASH_N_CHAR || _cur == SLASH_R_CHAR) throw _err("Unexpected newline in bin literal");
      s.writeCharCode(_cur);
      consume();
    }
    consume();
    return new HBin(s.toString());
  }

  HVal _readCoordVal() {
    if (_cur < 0) throw _err("Expected '(' after Coord");
    consume();
    StringBuffer s = new StringBuffer("C(");
    while (_cur != RIGHT_CURLY_CHAR)
    {
      if (_cur < 0) throw _err("Unexpected end of coord literal");
      if (_cur == SLASH_N_CHAR || _cur == SLASH_R_CHAR) throw _err("Unexpected newline in coord literal");
      s.writeCharCode(_cur);
      consume();
    }
    consume();
    s.write(")");
    return new HCoord.fromString(s.toString());
  }

  HVal _readNumVal() {
    // parse numeric part
    StringBuffer s = new StringBuffer();
    s.writeCharCode(_cur);
    consume();
    while (_isDigit(_cur) || _cur == PERIOD_CHAR || _cur == UNDERSCORE_CHAR) {
      if (_cur != UNDERSCORE_CHAR) s.writeCharCode(_cur);
      consume();
      if (_cur == e_CHAR || _cur == E_CHAR) {
        if (_peek == MINUS_CHAR || _peek == PLUS_CHAR || _isDigit(_peek)) {
          s.writeCharCode(_cur); consume();
          s.writeCharCode(_cur); consume();
        }
      }
    }
    double val = double.parse(s.toString());

    // HDate - check for dash
    HDate date = null;
    HTime time = null;
    int hour = -1;
    if (_cur == MINUS_CHAR) {
      int year;
      try {
        year = int.parse(s.toString());
      } on Exception catch (e) {
        throw _err("Invalid year for date value: $s");
      }
      consume(); // dash
      int month = _readTwoDigits("Invalid digit for month in date value");
      if (_cur != MINUS_CHAR) throw _errChar("Expected '-' for date value");
      consume();
      int day = _readTwoDigits("Invalid digit for day in date value");
      date = new HDate(year, month, day);

      // check for 'T' date time
      if (_cur != T_CHAR) return date;

      // parse next two digits and drop down to HTime parsing
      consume();
      hour = _readTwoDigits("Invalid digit for hour in date time value");
    }

    //HTime - check for colon
    if (_cur == COLON_CHAR) {
      //hour (may have been parsed already in date time)
      if (hour < 0) {
        if (s.length != 2) throw _err("Hour must be two digits for time value: $s");
        try {
          hour = int.parse(s.toString());
        } on Exception catch (e) {
          throw _err("Invalid hour for time value: $s");
        }
      }
      consume(); // colon
      int min = _readTwoDigits("Invalid digit for minute in time value");
      if (_cur != COLON_CHAR) throw _errChar("Expected ':' for time value");
      consume();
      int sec = _readTwoDigits("Invalid digit for seconds in time value");
      int ms = 0;
      if (_cur == PERIOD_CHAR) {
        consume();
        int places = 0;
        while (_isDigit(_cur)) {
          ms = (ms * 10) + (_cur - ZERO_CHAR);
          consume();
          places++;
        } switch (places) {
          case 1: ms *= 100; break;
          case 2: ms *= 10; break;
          case 3: break;
          default: throw _err("Too many digits for milliseconds in time value");
        }
      }
      time = new HTime(hour, min, sec, ms);
      if (date == null) return time;
    }

    //HDateTime (if we have date and time)
    bool zUtc = false;
    if (date != null) {
      //timezone offset "Z" or "-/+hh:mm"
      int tzOffset = 0;
      if (_cur == Z_CHAR) {
        consume();
        zUtc = true;
      } else {
        bool neg = _cur == MINUS_CHAR;
        if (_cur != MINUS_CHAR && _cur != PLUS_CHAR) throw _errChar("Expected -/+ for timezone offset");
        consume();
        int tzHours = _readTwoDigits("Invalid digit for timezone offset");
        if (_cur != COLON_CHAR) throw _errChar("Expected colon for timezone offset");
        consume();
        int tzMins  = _readTwoDigits("Invalid digit for timezone offset");
        tzOffset = (tzHours * 3600) + (tzMins * 60);
        if (neg) tzOffset = -tzOffset;
      }

      //timezone name
      HTimeZone tz;
      if (_cur != SPACE_CHAR) {
        if (!zUtc)
          throw _errChar("Expected space between timezone offset and name");
        else
          tz = HTimeZone.UTC;
      } else if (zUtc && !(A_CHAR <= _peek && _peek <= Z_CHAR)) {
        tz = HTimeZone.UTC;
      } else {
        consume();
        StringBuffer tzBuf = new StringBuffer();
        if (!_isTz(_cur)) throw _errChar("Expected timezone name");
        while (_isTz(_cur)) { tzBuf.writeCharCode(_cur); consume(); }
        tz = new HTimeZone(tzBuf.toString(), tzOffset);
      }
      return new HDateTime(date, time, tz);
    }

    // if we have unit, parse that
    String unit = null;
    if (_isUnit(_cur)) {
      s = new StringBuffer();
      while (_isUnit(_cur)) { s.writeCharCode(_cur); consume(); }
      unit = s.toString();
    }

    return new HNum(val, unit);
  }

  int _readTwoDigits(String errMsg) {
    if (!_isDigit(_cur)) throw _errChar(errMsg);
    int tens = (_cur - ZERO_CHAR) * 10;
    consume();
    if (!_isDigit(_cur)) throw _errChar(errMsg);
    int val = tens + (_cur - ZERO_CHAR);
    consume();
    return val;
  }

  HVal _readRefVal() {
    consume(); // opening @
    StringBuffer s = new StringBuffer();
    while (HRef.isIdChar(_cur)) {
      if (_cur < 0) throw _err("Unexpected end of ref literal");
      if (_cur == SLASH_N_CHAR || _cur == SLASH_R_CHAR) throw _err("Unexpected newline in ref literal");
      s.writeCharCode(_cur);
      consume();
    }
    _skipSpace();

    String dis = null;
    if (_cur == DOUBLE_QUOTE_CHAR) dis = _readStrLiteral();

    return new HRef(s.toString(), dis);
  }

  HVal _readStrVal() => new HStr(_readStrLiteral());

  String _readStrLiteral() {
    consume(); // opening quote
    StringBuffer s = new StringBuffer();
    while (_cur != DOUBLE_QUOTE_CHAR) {
      if (_cur < 0) throw _err("Unexpected end of str literal");
      if (_cur == SLASH_N_CHAR || _cur == SLASH_R_CHAR) throw _err("Unexpected newline in str literal");
      if (_cur == BACK_SLASH_CHAR) {
        s.writeCharCode(_readEscChar());
      } else {
        s.writeCharCode(_cur);
        consume();
      }
    }
    consume(); // closing quote
    return s.toString();
  }

  int _readEscChar() {
    consume();  // back slash

    // check basics
    switch (_cur) {
      case b_CHAR:   consume(); return SLASH_B_CHAR;
      case f_CHAR:   consume(); return SLASH_F_CHAR;
      case n_CHAR:   consume(); return SLASH_N_CHAR;
      case r_CHAR:   consume(); return SLASH_R_CHAR;
      case t_CHAR:   consume(); return SLASH_T_CHAR;
      case DOUBLE_QUOTE_CHAR:   consume(); return DOUBLE_QUOTE_CHAR;
      case DOLLAR_CHAR:   consume(); return DOLLAR_CHAR;
      case BACK_SLASH_CHAR:  consume(); return BACK_SLASH_CHAR;
    }

    // check for uxxxx
    if (_cur == u_CHAR) {
      consume();
      int n3 = _toNibble(_cur); consume();
      int n2 = _toNibble(_cur); consume();
      int n1 = _toNibble(_cur); consume();
      int n0 = _toNibble(_cur); consume();
      return (n3 << 12) | (n2 << 8) | (n1 << 4) | (n0);
    }

    throw _err("Invalid escape sequence: \\${new String.fromCharCode(_cur)}");
  }

  int _toNibble(int c) {
    if (ZERO_CHAR <= c && c <= NINE_CHAR) return c - ZERO_CHAR;
    if (a_CHAR <= c && c <= f_CHAR) return c - a_CHAR + 10;
    if (A_CHAR <= c && c <= F_CHAR) return c - A_CHAR + 10;
    throw _errChar("Invalid hex char");
  }

  HVal _readUriVal() {
    consume(); // opening backtick
    StringBuffer s = new StringBuffer();

    while (true) {
      if (_cur < 0) throw _err("Unexpected end of uri literal");
      if (_cur == SLASH_N_CHAR || _cur == SLASH_R_CHAR) throw _err("Unexpected newline in uri literal");
      if (_cur == GRAVE_ACCENT_CHAR) break;
      if (_cur == BACK_SLASH_CHAR) {
        switch (_peek) {
          case COLON_CHAR: case SLASH_CHAR: case QUESTION_MARK_CHAR: case NUMBER_SIGN_CHAR:
          case LEFT_SQUARE_BRACKET_CHAR: case RIGHT_SQUARE_BRACKET_CHAR: case DOG_CHAR: case BACK_SLASH_CHAR:
          case AMPERSAND_CHAR: case EQUALS_SIGN_CHAR: case SEMICOLON_CHAR:
            s.writeCharCode(_cur);
            s.writeCharCode(_peek);
            consume();
            consume();
            break;
          case GRAVE_ACCENT_CHAR:
            s.writeCharCode(GRAVE_ACCENT_CHAR);
            consume();
            consume();
            break;
          default:
            if (_peek == u_CHAR || _peek == BACK_SLASH_CHAR) s.writeCharCode(_readEscChar());
            else throw _err("Invalid URI escape sequence \\${new String.fromCharCode(_peek)}");
            break;
        }
      } else {
        s.writeCharCode(_cur);
        consume();
      }
    }
    consume(); // closing backtick
    return new HUri(s.toString());
  }


//////////////////////////////////////////////////////////////////////////
// Char Reads
//////////////////////////////////////////////////////////////////////////

  ParseError _errChar(String msg){
    if (_cur < 0)
      msg += " (end of stream)";
    else {
      msg += " (char=0x" + _cur.toRadixString(16);
      if (_cur >= SPACE_CHAR) msg += " '" + new String.fromCharCode(_cur) + "'";
      msg += ")";
    }
    return _err(msg);
  }

  ParseError _err(String msg) => new ParseError(msg + " [Line $_lineNum]");

  void _skipSpace() {
    while (_cur == SPACE_CHAR || _cur == SLASH_T_CHAR) consume();
  }

  void _consumeNewline() {
    if (_cur != SLASH_N_CHAR) throw _errChar("Expecting newline");
    consume();
  }

  void consume() {
    try {
      _cur  = _peek;
      _peek = _in.read();
      if (_cur == SLASH_N_CHAR) _lineNum++;
    } on StateError catch (error) {
      throw new ParseError(error.message);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Char Types
//////////////////////////////////////////////////////////////////////////

  static bool _isDigit(int c)   => c > 0 && c < 128 && (_charTypes[c] & _DIGIT) != 0;
  static bool _isAlpha(int c)   => c > 0 && c < 128 && (_charTypes[c] & _ALPHA) != 0;
  static bool _isUnit(int c)    => c > 0 && (c >= 128 || (_charTypes[c] & _UNIT) != 0);
  static bool _isTz(int c)      => c > 0 && c < 128 && (_charTypes[c] & _TZ) != 0;
  static bool _isIdStart(int c) => c > 0 && c < 128 && (_charTypes[c] & _ID_START) != 0;
  static bool _isId(int c)      => c > 0 && c < 128 && (_charTypes[c] & _ID) != 0;

  static final int _DIGIT    = 0x01;
  static final int _ALPHA_LO = 0x02;
  static final int _ALPHA_UP = 0x04;
  static final int _ALPHA    = _ALPHA_UP | _ALPHA_LO;
  static final int _UNIT     = 0x08;
  static final int _TZ       = 0x10;
  static final int _ID_START = 0x20;
  static final int _ID       = 0x40;

  static final List<int> _charTypes = (){
    List<int> result = new List<int>.filled(128, 0);

    for (int i=ZERO_CHAR; i<=NINE_CHAR; ++i) result[i] = (_DIGIT | _TZ | _ID);
    for (int i=a_CHAR; i<=z_CHAR; ++i) result[i] = (_ALPHA_LO | _UNIT | _TZ | _ID_START | _ID);
    for (int i=A_CHAR; i<=Z_CHAR; ++i) result[i] = (_ALPHA_UP | _UNIT | _TZ | _ID);
    result[PERCENT_CHAR] = _UNIT;
    result[UNDERSCORE_CHAR] = _UNIT | _TZ | _ID;
    result[SLASH_CHAR] = _UNIT;
    result[DOLLAR_CHAR] = _UNIT;
    result[MINUS_CHAR] = _TZ;
    result[PLUS_CHAR] = _TZ;

    return result;
  }();

  //////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final StringReader _in;
  int _cur = 0;
  int _peek = 0;
  int _version;
  int _lineNum = 1;
  bool _isFilter = false;
}