///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.dict_test;

import 'package:unittest/unittest.dart';
import 'package:haystack/core.dart';

main () {
  group('Test empty HDict', () {
    HDict tags = new HDictBuilder().toDict();

    test('equality', () {
      expect(identical(tags, HDict.EMPTY), equals(true));
      expect(tags == HDict.EMPTY, equals(true));
    });

    test('size', () {
      expect(tags.size, equals(0));
      expect(tags.isEmpty(), equals(true));
    });

    test('missing tag', () {
      expect(tags.has("foo"), equals(false));
      expect(tags.missing("foo"), equals(true));
      expect(tags.get("foo", checked:false), equals(null));
      expect(() => tags.get("foo"), throwsA(new isInstanceOf<UnknownNameError>()));
      expect(() => tags.get("foo", checked:true), throwsA(new isInstanceOf<UnknownNameError>()));
    });
  });

  test('Test isTagName', () {
    expect(HDict.isTagName(""), equals(false));
    expect(HDict.isTagName("A"), equals(false));
    expect(HDict.isTagName(" "), equals(false));
    expect(HDict.isTagName("a"), equals(true));
    expect(HDict.isTagName("a_B_19"), equals(true));
    expect(HDict.isTagName("a b"), equals(false));
    expect(HDict.isTagName("a\u0128"), equals(false));
    expect(HDict.isTagName("a\u0129x"), equals(false));
    expect(HDict.isTagName("a\uabcdx"), equals(false));
  });

  group('Test basics', () {
    HDict tags = new HDictBuilder()
    .add("id", new HRef("aaaa-bbbb"))
    .add("site")
    .add("geoAddr", "Richmond, Va")
    .add("area", 1200, "ft")
    .add("date", new HDate(2000, 12, 3))
    .toDict();

    test('size', () {
      expect(tags.size, equals(5));
      expect(tags.isEmpty(), equals(false));
    });

    test('configured tag', () {
      expect(tags.get("id"),      equals(new HRef("aaaa-bbbb")));
      expect(tags.get("site"),    equals(HMarker.VAL));
      expect(tags.get("geoAddr"), equals(new HStr("Richmond, Va")));
      expect(tags.get("area"),    equals(new HNum(1200, "ft")));
      expect(tags.get("date"),    equals(new HDate(2000, 12, 3)));
    });

    test('missing tag', () {
      expect(tags.has("foo"), equals(false));
      expect(tags.missing("foo"), equals(true));
      expect(tags.get("foo", checked:false), equals(null));
      expect(() => tags.get("foo"), throwsA(new isInstanceOf<UnknownNameError>()));
      expect(() => tags.get("foo", checked:true), throwsA(new isInstanceOf<UnknownNameError>()));
    });
  });

  test('Test equality', () {
    HDict a = new HDictBuilder().add("x").toDict();

    expect(a == new HDictBuilder().add("x").toDict(), equals(true));
    expect(a == new HDictBuilder().add("x", 3).toDict(), equals(false));
    expect(a == new HDictBuilder().add("y").toDict(), equals(false));
    expect(a == new HDictBuilder().add("x").add("y").toDict(), equals(false));

    a = new HDictBuilder().add("x").add("y", "str").toDict();
    expect(a == new HDictBuilder().add("x").add("y", "str").toDict(), equals(true));
    expect(a == new HDictBuilder().add("y", "str").add("x").toDict(), equals(true));
    expect(a == new HDictBuilder().add("x", "str").add("y", "str").toDict(), equals(false));
    expect(a == new HDictBuilder().add("x").add("y", "strx").toDict(), equals(false));
    expect(a == new HDictBuilder().add("y", "str").toDict(), equals(false));
    expect(a == new HDictBuilder().add("x").toDict(), equals(false));
    expect(a == new HDictBuilder().add("x").add("yy", "str").toDict(), equals(false));
  });

  test('Test zinc', () {
    verifyZinc("", HDict.EMPTY);
    verifyZinc("foo_12",
        new HDictBuilder().add("foo_12").toDict());
    verifyZinc("fooBar:123ft",
        new HDictBuilder().add("fooBar", 123, "ft").toDict());
    verifyZinc("dis:\"Bob\" bday:1970-06-03 marker",
        new HDictBuilder().add("dis", "Bob").add("bday", new HDate(1970,6,3)).add("marker").toDict());
    verifyZinc("dis  :  \"Bob\"  bday : 1970-06-03  marker",
        new HDictBuilder().add("dis", "Bob").add("bday", new HDate(1970,6,3)).add("marker").toDict());
  });

  test('Test dis', () {
    expect(new HDictBuilder().add("id", new HRef("a")).toDict().dis(), equals("a"));
    expect(new HDictBuilder().add("id", new HRef("a", "b")).toDict().dis(), equals("b"));
    expect(new HDictBuilder().add("id", new HRef("a")).add("dis", "d").toDict().dis(), equals("d"));
  });

}

void verifyZinc(String s, HDict tags) {
  HDict x = new HZincReader(s).readDict();
  if (tags.size <= 1) expect(tags.toZinc(), equals(s));
  expect(x, equals(tags));
}